/*
 * Zorin Taskbar: A taskbar extension for the Zorin Desktop environment.
 * Copyright (C) 2019 Zorin OS Technologies Ltd.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Credits:
 * This file is based on code from the Dash to Dock extension by micheleg
 * and code from the Dash to Panel extension by Jason DeRose (as at
 * commit b6094fdaec89349cc6f3e0da887d19fdf3db1c60)
 * Some code was also adapted from the upstream Gnome Shell source code.
 */

const Me = imports.misc.extensionUtils.getCurrentExtension();
const Utils = Me.imports.utils;

const Clutter = imports.gi.Clutter;
const Lang = imports.lang;
const Main = imports.ui.main;
const Shell = imports.gi.Shell;
const Mainloop = imports.mainloop;

const Meta = imports.gi.Meta;

var OVERLAY_TIMEOUT = 750;
var SHORTCUT_TIMEOUT = 2000;
var HOTKEY_PREFIX = 'Super';

var zorinTaskbarOverview = new Lang.Class({
    Name: 'ZorinTaskbar.Overview',

    _numHotkeys: 10,

    _init: function(settings) {
        this._zorinTaskbarSettings = settings;
    },

    enable : function(panel) {
        this._panel = panel;
        this.taskbar = panel.taskbar;

        this._signalsHandler = new Utils.GlobalSignalsHandler();

        this._optionalHotKeys();
        this._optionalNumberOverlay();
        this._toggleDash(false);
    },

    disable: function () {
        this._toggleDash(true);

        // reset stored icon size  to the default dash
        Main.overview.dashIconSize = Main.overview._controls.dash.iconSize;

        // Remove key bindings
        this._disableHotKeys();
        this._disableExtraShortcut();
    },

    _toggleDash: function(visible) {
        // To hide the dash, set its width to 1, so it's almost not taken into account by code
        // calculaing the reserved space in the overview. The reason to keep it at 1 is
        // to allow its visibility change to trigger an allocaion of the appGrid which
        // in turn is triggergin the appsIcon spring animation, required when no other
        // actors has this effect, i.e in horizontal mode and without the workspaceThumnails
        // 1 static workspace only)

        let visibilityFunc = visible ? 'show' : 'hide';
        let width = visible ? -1 : 1;
        
        Main.overview._controls.dash.actor[visibilityFunc]();
        Main.overview._controls.dash.actor.set_width(width);

        // This force the recalculation of the icon size
        Main.overview._controls.dash._maxHeight = -1;
    },

    // Hotkeys
    _activateApp: function(appIndex) {
        let seenApps = {};
        let apps = [];
        
        this.taskbar._getAppIcons().forEach(function(appIcon) {
            if (!seenApps[appIcon.app]) {
                apps.push(appIcon);
            }

            seenApps[appIcon.app] = (seenApps[appIcon.app] || 0) + 1;
        });

        if (appIndex < apps.length) {
            let appIcon = apps[appIndex];
            let windowCount = !appIcon.window ? appIcon._nWindows : seenApps[appIcon.app];

            if (windowCount > 1) {
                let windowIndex = 0;
                let currentWindow = appIcon.window;
                let hotkeyPrefix = HOTKEY_PREFIX;
                let shortcutKeys = (hotkeyPrefix == 'Super' ? [Clutter.Meta_L] : [Clutter.Meta_L, Clutter.Alt_L]).concat([Clutter['KEY_' + (appIndex + 1)]]);
                let currentKeys = shortcutKeys.slice();
                let setFocus = windowIndex => global.stage.set_key_focus(appIcon.windowPreview._previewBox.box.get_children()[windowIndex]);
                let capturedEventId = global.stage.connect('captured-event', (actor, event) => {
                    if (event.type() == Clutter.EventType.KEY_PRESS) {
                        let pressedKey = event.get_key_symbol();

                        if (shortcutKeys.indexOf(pressedKey) >= 0 && currentKeys.indexOf(pressedKey) < 0) {
                            currentKeys.push(pressedKey);

                            if (shortcutKeys.length === currentKeys.length) {
                                windowIndex = windowIndex < windowCount - 1 ? windowIndex + 1 : 0;
                                setFocus(windowIndex);
                            }
                        } 
                    } else if (event.type() == Clutter.EventType.KEY_RELEASE) {
                        let keyIndex = currentKeys.indexOf(event.get_key_symbol());

                        if (keyIndex >= 0) {
                            currentKeys.splice(keyIndex, 1);
                        }
                    }

                    return Clutter.EVENT_PROPAGATE;
                });
                let hotkeyOpenStateChangedId = appIcon.windowPreview.connect('open-state-changed', (menu, isOpen) => {
                    if (!isOpen) {
                        global.stage.disconnect(capturedEventId);
                        appIcon.windowPreview.disconnect(hotkeyOpenStateChangedId);
                        appIcon.windowPreview._previewBox._resetPreviews();
                        appIcon.window = currentWindow;
                    }
                });

                appIcon.window = null;
                appIcon.windowPreview.popup();
                appIcon.menuManagerWindowPreview._onMenuOpenState(appIcon.windowPreview, true);
                setFocus(windowIndex);
            } else {
                // Activate with button = 1, i.e. same as left click
                let button = 1;
                appIcon.activate(button, true);
            }
        }
    },

    _optionalHotKeys: function() {
        this._hotKeysEnabled = false;
        this._enableHotKeys();
    },

    _enableHotKeys: function() {
        if (this._hotKeysEnabled)
            return;

        // Setup keyboard bindings for taskbar elements
        let keys = ['app-hotkey-', 'app-shift-hotkey-', 'app-ctrl-hotkey-',  // Regular numbers
                    'app-hotkey-kp-', 'app-shift-hotkey-kp-', 'app-ctrl-hotkey-kp-']; // Key-pad numbers
        keys.forEach( function(key) {
            for (let i = 0; i < this._numHotkeys; i++) {
                let appNum = i;

                if (!Main.wm._allowedKeybindings[key + (i + 1)]) {
                    Main.wm.addKeybinding(key + (i + 1), this._zorinTaskbarSettings,
                                        Meta.KeyBindingFlags.NONE,
                                        Shell.ActionMode.NORMAL | Shell.ActionMode.OVERVIEW,
                                        Lang.bind(this, function() {
                                            this._activateApp(appNum);
                                            this._showOverlay();
                                        }));
                }
            }
        }, this);

        this._hotKeysEnabled = true;
    },

    _disableHotKeys: function() {
        if (!this._hotKeysEnabled)
            return;

        let keys = ['app-hotkey-', 'app-shift-hotkey-', 'app-ctrl-hotkey-',  // Regular numbers
                    'app-hotkey-kp-', 'app-shift-hotkey-kp-', 'app-ctrl-hotkey-kp-']; // Key-pad numbers
        keys.forEach( function(key) {
            for (let i = 0; i < this._numHotkeys; i++) {
                if (Main.wm._allowedKeybindings[key + (i + 1)]) {
                    Main.wm.removeKeybinding(key + (i + 1));
                }
            }
        }, this);

        this._hotKeysEnabled = false;

        this.taskbar.toggleNumberOverlay(false);
    },

    _optionalNumberOverlay: function() {
        this._enableExtraShortcut();
    },

    _enableExtraShortcut: function() {
        if (!Main.wm._allowedKeybindings['shortcut']) {
            Main.wm.addKeybinding('shortcut', this._zorinTaskbarSettings,
                                  Meta.KeyBindingFlags.NONE,
                                  Shell.ActionMode.NORMAL | Shell.ActionMode.OVERVIEW,
                                  Lang.bind(this, function() {
                                      this._showOverlay(true);
                                  }));
        }
    },

    _disableExtraShortcut: function() {
        if (Main.wm._allowedKeybindings['shortcut']) {
            Main.wm.removeKeybinding('shortcut');
        }
    },

    _showOverlay: function(overlayFromShortcut) {
        // Restart the counting if the shortcut is pressed again
        if (this._numberOverlayTimeoutId) {
            Mainloop.source_remove(this._numberOverlayTimeoutId);
            this._numberOverlayTimeoutId = 0;
        }

        this.taskbar.toggleNumberOverlay(true);

        let timeout = OVERLAY_TIMEOUT;

        if (overlayFromShortcut) {
            timeout = SHORTCUT_TIMEOUT;
        }

        // Hide the overlay/dock after the timeout
        this._numberOverlayTimeoutId = Mainloop.timeout_add(timeout, Lang.bind(this, function() {
                this._numberOverlayTimeoutId = 0;
                this.taskbar.toggleNumberOverlay(false);
        }));
    }
});
